#include "texture_manager.hpp"
#include "exceptions.hpp"
#include "vulkan_instance.hpp"
#include "vulkan_util.hpp"

#include <glm/detail/qualifier.hpp>
#include <glm/glm.hpp>
#include <gz-util/exceptions.hpp>
#include <gz-util/string/to_string.hpp>

namespace gz::vlk {
	std::string TextureInfo::toString() const {
        return "<atlas: " + gz::toString(atlas) + ", tl: " + gz::toString(texCoordTopLeft) + ", br: " + gz::toString(texCoordBottomRight) + ">";
	}


    TextureManager::TextureManager(VulkanInstance& instance)
        : vk(instance)
    {
        LogCreateInfo logCI{
        	.logfile = "textureManager.log",
        	.storeLog = false,
        	.prefix = "Texture",
        	.prefixColor = Color::GREEN,
        	.timeColor = settings::VULKAN_MESSAGE_TIME_COLOR,
        };
        tLog = Log(std::move(logCI));

        vk.registerCleanupCallback(std::bind(&TextureManager::cleanup, this));
        addAtlas(24, 24, 24, 24);
        createDescriptorResources();
        VulkanInstance::registerObjectUsingVulkan("TextureManager",
                &descriptorSetLayout, &descriptorPool, &descriptorSet);
    }


    void TextureManager::cleanup() {
        tLog.log0("cleanup: cleaning up");
        vk.getDevice().destroyDescriptorSetLayout(descriptorSetLayout, NO_ALLOC);
        vk.getDevice().destroyDescriptorPool(descriptorPool, NO_ALLOC);
        for (auto& atlas : atlases) {
            atlas.cleanup();
        }
    }


	TextureAtlasIndex TextureManager::addAtlas(uint16_t slotWidth, uint16_t slotHeight, uint16_t slotCountX, uint16_t slotCountY, vk::Format imageFormat) {
        TextureAtlasIndex index = atlases.size();
        atlases.emplace_back(TextureAtlas(vk, slotWidth, slotHeight, slotCountX, slotCountY, imageFormat));
        // add special "atlas_X" texture
        textures.insert( { std::string("atlas_") + gz::toString(index), index });
        textureInfos.emplace_back(TextureInfo{ .atlas = index, .texCoordTopLeft = glm::vec2(0, 0), .texCoordBottomRight = glm::vec2(1, 1) });
        return index;
	}


	void TextureManager::isValidTexture(const Texture& texture) const {
        if (texture < textureInfos.size()) {
            return;
        }		
        else {
            throw InvalidArgument("Texture handle '" + gz::toString(texture) + "' is invalid.", "TextureManager::isValidTexture");
        }
	}


    Texture TextureManager::getTexture(const std::string& textureName) {
        if (!textures.contains(textureName)) {
            return loadTextureFromFile(textureName);
        }
        else {
            return textures.at(textureName);
        }
	}


	void TextureManager::setTexCoords(const Texture& texture, glm::vec2& texCoords) {
        isValidTexture(texture);
        TextureInfo& texI = textureInfos.at(texture);
        tLog.log0("setTexCoords", texCoords, "textureInfo: topleft", texI.texCoordTopLeft, "botright", texI.texCoordBottomRight);
        texCoords = texI.texCoordTopLeft + texCoords * (texI.texCoordBottomRight - texI.texCoordTopLeft);
	}


//
// ATLASES
//
	const TextureAtlas& TextureManager::getTextureAtlas(const Texture& texture) const {
        isValidTexture(texture);
        return atlases.at(textureInfos.at(texture).atlas);
	}
	uint32_t TextureManager::getTextureAtlasIndex(const Texture& texture) const {
        isValidTexture(texture);
        return textureInfos.at(texture).atlas;
	}

	Texture TextureManager::loadTextureFromFile(const std::string& textureName) {
        int textureWidth, textureHeight, textureChannels;
        stbi_uc* pixels = stbi_load(("textures/" + textureName).c_str(), &textureWidth, &textureHeight, &textureChannels, STBI_rgb_alpha);
        if (!pixels) {
            throw Exception("Failed to load texture: textures/" + textureName, "TextureManager::loadTexture");
        }
        Texture texture = textureInfos.size();
        textureInfos.emplace_back();
        TextureInfo& texI = textureInfos.back();
        texI.atlas = 0;
        static constexpr uint8_t BYTES_PER_PIXEL = 4;
        auto texCoords = atlases.at(0).addTexture(pixels, textureWidth, textureHeight, BYTES_PER_PIXEL);
        stbi_image_free(pixels);
        tLog.log0("TextureManager::loadTextureFromFile: TexCoords of new image:", texCoords.first, texCoords.second);
        texI.texCoordTopLeft = std::move(texCoords.first),
        texI.texCoordBottomRight = std::move(texCoords.second),
        tLog.log0("TextureManager::loadTextureFromFile: After loading:", atlases.at(0));
        tLog.log0("loadTextureFromFile: textureInfos", textureInfos);
        return texture;
	}


//
// DESCRIPTOR SET
//
    void TextureManager::createDescriptorResources() {
        // 1) layout
        // combined image sampler
        std::vector<vk::DescriptorSetLayoutBinding> samplerLayoutBindings(atlases.size(), {
                .binding = 0,
                .descriptorType = vk::DescriptorType::eCombinedImageSampler,
                .descriptorCount = 1,
                .stageFlags = vk::ShaderStageFlagBits::eFragment,
                /* .pImmutableSamplers = nullptr, */
            }
        );
        vk.createDescriptorSetLayout(samplerLayoutBindings, descriptorSetLayout);

        // 2) pool
        vk::DescriptorPoolSize poolSize { 
            .type = vk::DescriptorType::eCombinedImageSampler,
            .descriptorCount = vk.getMaxFramesInFlight(),
        };
        vk.createDescriptorPool({ poolSize }, vk.getMaxFramesInFlight(), descriptorPool);

        // 3) Set
        vk.createDescriptorSet(descriptorSetLayout, descriptorPool, descriptorSet);

        // 4) configure sets
        std::vector<vk::DescriptorImageInfo> imageInfos;
        for (auto it = atlases.cbegin(); it != atlases.cend(); it++) {
            imageInfos.emplace_back();
            imageInfos.back().imageLayout = vk::ImageLayout::eShaderReadOnlyOptimal;
            imageInfos.back().imageView = it->getTextureImageView();
            imageInfos.back().sampler = it->getTextureSampler();
        }

        vk::WriteDescriptorSet descriptorW {
            .dstSet = descriptorSet,
            .dstBinding = 0,
            .dstArrayElement = 0,
            .descriptorCount = 1,
            .descriptorType = vk::DescriptorType::eCombinedImageSampler,
        };
        descriptorW.setImageInfo(imageInfos);
        
        // 1 write, no copies
        vk.getDevice().updateDescriptorSets(descriptorW, nullptr);
        tLog.log0("createDescriptorSets: Created descriptor sets.");
    }

}  // namespace gz::vlk
